ACTIONS = {'NORTH': (-1, 0), 'SOUTH': (1, 0), 'WEST': (0, -1), 'EAST': (0, 1)}
MAZE_OBJECTS = {'EMPTY', 'OBSTACLE', 'GOLD', 'START'}

class State:
    """State of an agent in a maze
    State variables:
    gold: 2-tuple, (r,c) maze coordinate of the Gold
    step: int, number of steps since the start of the simulation
    position: 2-tuple, (r,c) coordinate of the current position of the agent
    vision: dictionary, number of cells to the closest obstacle or edge of the maze in each direction from ACTIONS.keys().
    """

    def __init__(self, vision, position):
        self.position = position
        self.vision = vision


class MazeKeeper:
    """
    Handles interaction between the agent and the maze.
    Takes agents actions and returns agents new state after executing the action
    """

    def __init__(self, layout):
        self.layout = layout
        self.agent_position = self._get_unique_cell(layout, 'START')
        self.gold_position = self._get_unique_cell(layout, 'GOLD')
        self._movable_positions = self._get_cell_positions(self.layout, ['EMPTY', 'GOLD', 'START'])
        self.finished = False

    @staticmethod
    def _get_cell_positions(layout, cell_types):
        positions = []
        for r, row in enumerate(layout):
            for c, cell in enumerate(row):
                if cell in cell_types:
                    positions.append((r, c))
        return positions

    @staticmethod
    def _get_unique_cell(layout, cell_type):
        cell = MazeKeeper._get_cell_positions(layout, [cell_type])
        if len(cell) == 1:
            return cell[0]
        else:
            raise Exception('Maze layout contains more than one {0} or does not contain {0} at all'.format(cell_type))

    @staticmethod
    def add_tuples(c1, c2):
        """Elementwise add for two tuples of the same length"""
        return tuple(v1 + v2 for v1, v2 in zip(c1, c2))

    def _agent_vision(self):
        """
        Returns vision of the agent from current position
        :return dictionary:
        key is direction from ACTIONS.keys(),
        value is number of cells to the closest obstacle or edge in given diraction from the current position of the agent.
        """
        vision = {}
        for direction, shift in ACTIONS.items():
            vision[direction] = -1
            tmp_position = self.agent_position
            while tmp_position in self._movable_positions:
                tmp_position = self.add_tuples(tmp_position, shift)
                vision[direction] += 1
        return vision

    def state(self):
        """Returns current state of the agent"""
        if self.agent_position == self.gold_position:
            self.finished = True
        return State(vision=self._agent_vision(),
                     position=self.agent_position)

    def agent_move(self, action):
        """
        Executes action of the agent.
        If the agent tries to move into an obstacle or edge of the maze, he stays at the same position.
        :return State: new state of the agent after executing the move.
        """
        new_position = self.add_tuples(self.agent_position, ACTIONS[action])
        if new_position in self._movable_positions:
            self.agent_position = new_position
        return self.state()


if __name__ == '__main__':
    E, O, G, S = 'EMPTY', 'OBSTACLE', 'GOLD', 'START'

    SAMPLE_LAYOUT = [[E, O, E, E, E],
                     [E, O, E, O, E],
                     [S, O, E, O, G],
                     [E, O, E, O, E],
                     [E, E, E, E, E]]

    maze = MazeKeeper(SAMPLE_LAYOUT)
    print(maze._get_cell_positions(SAMPLE_LAYOUT, S))
    print(maze._agent_vision())
